import sys
from qtpy.QtCore import QThread
from easydict import EasyDict as edict
from pymodaq.utils.daq_utils import ThreadCommand, getLineInfo
from pymodaq.utils.data import DataFromPlugins
from pymodaq.control_modules.viewer_utility_classes import DAQ_Viewer_base
from collections import OrderedDict
import numpy as np
import clr
from pymodaq.control_modules.viewer_utility_classes import comon_parameters


class DAQ_0DViewer_Kinesis_KPA101(DAQ_Viewer_base):
    """
        ==================== ========================
        **Attributes**        **Type**
        *data_grabed_signal*  instance of Signal
        *VISA_rm*             ResourceManager
        *com_ports*           
        *params*              dictionnary list
        *keithley*
        *settings*
        ==================== ========================
    """
    ##checking VISA ressources

    _controller_units = 'V'
    kinesis_path = 'C:\\Program Files\\Thorlabs\\Kinesis'

    try:
        sys.path.append(kinesis_path)
        clr.AddReference("Thorlabs.MotionControl.DeviceManagerCLI")
        clr.AddReference("Thorlabs.MotionControl.KCube.PositionAlignerCLI")
        import Thorlabs.MotionControl.DeviceManagerCLI as Device
        import Thorlabs.MotionControl.KCube.PositionAlignerCLI as PosAligner
        Device.DeviceManagerCLI.BuildDeviceList()
        # %%
        serialnumbers = [str(ser) for ser in
                         Device.DeviceManagerCLI.GetDeviceList(PosAligner.KCubePositionAligner.DevicePrefix)]
        # %%

    except:
        serialnumbers=[]

    params = comon_parameters+[
            {'title': 'Kinesis library:', 'name': 'kinesis_lib', 'type': 'browsepath', 'value': kinesis_path},
            {'title': 'Serial number:', 'name': 'serial_number', 'type': 'list', 'limits': serialnumbers},
            {'title': 'Device:', 'name': 'device_name', 'type': 'str', 'value': ''},
            {'title': 'Polling time (ms):', 'name': 'polling_time', 'type': 'int', 'value': 250},
            ]

    def __init__(self, parent=None, params_state=None):
        super().__init__(parent, params_state)
        self.controller = None

    def ini_detector(self, controller=None):
        """
            Initialisation procedure of the detector.

            Returns
            -------

                The initialized status.

            See Also
            --------
            daq_utils.ThreadCommand
        """
        self.status.update(edict(initialized=False, info="", x_axis=None, y_axis=None, controller=None))
        try:

            if self.settings.child(('controller_status')).value() == "Slave":
                if controller is None: 
                    raise Exception('no controller has been defined externally while this detector is a slave one')
                else:
                    self.controller = controller
            else:
                self.Device.DeviceManagerCLI.BuildDeviceList()
                serialnumbers = self.Device.DeviceManagerCLI.GetDeviceList(self.PosAligner.KCubePositionAligner.DevicePrefix)
                ser_bool = self.settings.child(('serial_number')).value() in serialnumbers
                if ser_bool:
                    self.controller = self.PosAligner.KCubePositionAligner.CreateKCubePositionAligner(
                        self.settings.child(('serial_number')).value())
                    self.controller.Connect(self.settings.child(('serial_number')).value())
                    if not self.controller.IsSettingsInitialized():
                        self.controller.WaitForSettingsInitialized(5000)
                    self.controller.StartPolling(self.settings.child(('polling_time')).value())
                    self.emit_status(ThreadCommand('update_main_settings', [['wait_time'],
                                                            self.settings.child(('polling_time')).value(), 'value']))
                    QThread.msleep(500)
                    self.controller.EnableDevice()
                    QThread.msleep(500)
                    deviceInfo = self.controller.GetDeviceInfo()
                    self.settings.child(('device_name')).setValue(deviceInfo.Name)

                else:
                    raise Exception("Not valid serial number")

            self.status.initialized = True
            self.status.controller = self.controller
            return self.status

        except Exception as e:
            self.emit_status(ThreadCommand('Update_Status', [getLineInfo() + str(e), 'log']))
            self.status.info = getLineInfo() + str(e)
            self.status.initialized = False
            return self.status


    def commit_settings(self, param):
        """
            Activate the parameters changes in the hardware.

            =============== ================================= ============================
            **Parameters**   **Type**                         **Description**
            *param*         instance of pyqtgraph.parameter   The parameter to be checked.
            =============== ================================= ============================

            See Also
            --------
            daq_utils.ThreadCommand
        """
        try:
            if param.name() == 'kinesis_lib':
                try:
                    sys.path.append(param.value())
                    clr.AddReference("Thorlabs.MotionControl.DeviceManagerCLI")
                    clr.AddReference("Thorlabs.MotionControl.IntegratedStepperMotorsCLI")
                    clr.AddReference("Thorlabs.MotionControl.GenericMotorCLI")
                    import Thorlabs.MotionControl.IntegratedStepperMotorsCLI as Integrated
                    import Thorlabs.MotionControl.DeviceManagerCLI as Device
                    import Thorlabs.MotionControl.GenericMotorCLI as Generic
                    Device.DeviceManagerCLI.BuildDeviceList()
                    serialnumbers = [str(ser) for ser in
                                     Device.DeviceManagerCLI.GetDeviceList(Integrated.CageRotator.DevicePrefix)]

                except:
                    serialnumbers = []
                self.settings.child(('serial_number')).setOpts(limits=serialnumbers)

            elif param.name() == 'polling_time':
                self.controller.StopPolling()
                QThread.msleep(500)
                self.controller.StartPolling(self.settings.child(('polling_time')).value())
                QThread.msleep(500)
                self.emit_status(ThreadCommand('update_main_settings', [['wait_time'], param.value(), 'value']))


        except Exception as e:
            self.emit_status(ThreadCommand('Update_Status', [getLineInfo() + str(e), 'log']))

    def close(self):
        """
            close the current instance of Keithley viewer.
        """
        self.controller.DisableDevice()
        self.controller.StopPolling()
        self.controller.Disconnect(False)

    def grab_data(self, Naverage=1, **kwargs):
        """
            | Start new acquisition.
            | grab the current values with keithley profile procedure.
            | Send the data_grabed_signal once done.

            =============== ======== ===============================================
            **Parameters**  **Type**  **Description**
            *Naverage*      int       Number of values to average
            =============== ======== ===============================================
        """
        status = self.controller.Status
        data = [np.array([status.PositionDifference.X]), np.array([status.PositionDifference.Y])]
        data_intens = [np.array([status.Sum])]
        self.data_grabed_signal.emit([DataFromPlugins(name='KPA101 Positions', data=data, dim='Data0D',
                                                      labels=['X (V)', 'Y (V)'],),
                                      DataFromPlugins(name='KPA101 Intensity', data=data_intens, dim='Data0D',
                                                      labels=['Intensity'],)])


    def stop(self):
        """
            not implemented?
        """
        return ""



"""
PyMoDAQ plugin for thorlabs instruments based on the TLPM library allowing
remote control and monitoring of up to eight power and energy meters.
This software is compatible with our Power Meter Consoles and Interfaces (PM100A and PM101 Series),
Power and Energy Meter Consoles and Interfaces (PM100D, PM400, PM100USB, PM103 Series, and legacy PM200),
Position & Power Meter Interfaces (PM102 Series),
Wireless Power Meters (PM160, PM160T, and PM160T-HP),
and USB-Interface Power Meters (PM16 Series)

you have to install the Optical Monitor Software from Thorlabs to obtain the library

The installation should create (following the manual) an environment variable called either VXIPNPPATH64 or
VXIPNPPATH depending on your platform (32 or 64 bits) pointing to where the TLPM library is
(usually C:\Program Files\IVI Foundation\VISA)

This plugin is making use of the TLPM.py script provided by thorlabs. An alternative is to use the TLPMPowermeterInst
plugin using the Instrumental_lib package directly interfacing the C library with the nice Instrument wrapper
"""
from easydict import EasyDict as edict
from pymodaq.utils.daq_utils import ThreadCommand, getLineInfo
from pymodaq.utils.data import DataFromPlugins
from pymodaq.control_modules.viewer_utility_classes import DAQ_Viewer_base, main

import numpy as np
from pymodaq.control_modules.viewer_utility_classes import comon_parameters
from pymodaq_plugins_thorlabs.hardware.powermeter import CustomTLPM, DEVICE_NAMES


class DAQ_0DViewer_TLPMPowermeter(DAQ_Viewer_base):

    _controller_units = 'W'
    devices = DEVICE_NAMES

    params = comon_parameters + [
        {'title': 'Devices:', 'name': 'devices', 'type': 'list', 'limits': devices},
        {'title': 'Info:', 'name': 'info', 'type': 'str', 'value': '', 'readonly': True},
        {'title': 'Wavelength:', 'name': 'wavelength', 'type': 'float', 'value': 532.,},
        ]

    def __init__(self, parent=None, params_state=None):
        super().__init__(parent, params_state)


    def ini_detector(self, controller=None):
        """
            Initialisation procedure of the detector.

            Returns
            -------

                The initialized status.

            See Also
            --------
            daq_utils.ThreadCommand
        """
        self.status.update(edict(initialized=False, info="", x_axis=None, y_axis=None, controller=None))
        try:

            if self.settings.child(('controller_status')).value() == "Slave":
                if controller is None:
                    raise Exception('no controller has been defined externally while this detector is a slave one')
                else:
                    self.controller = controller
            else:
                index = DEVICE_NAMES.index(self.settings['devices'])
                self.controller = CustomTLPM()
                info = self.controller.infos.get_devices_info(index)
                self.controller.open_by_index(index)
                self.settings.child('info').setValue(str(info))

            self.settings.child('wavelength').setOpts(limits=self.controller.wavelength_range)
            self.controller.wavelength = self.settings.child('wavelength').value()
            self.settings.child('wavelength').setValue(self.controller.wavelength)

            self.status.initialized = True
            self.status.controller = self.controller
            self.status.info = str(info)
            return self.status

        except Exception as e:
            self.emit_status(ThreadCommand('Update_Status', [getLineInfo() + str(e), 'log']))
            self.status.info = getLineInfo() + str(e)
            self.status.initialized = False
            return self.status


    def commit_settings(self, param):
        """
        """
        if param.name() == 'wavelength':
            self.controller.wavelength = self.settings.child('wavelength').value()
            self.settings.child('wavelength').setValue(self.controller.wavelength)

    def close(self):
        """
            close the current instance of Keithley viewer.
        """
        self.controller.close()

    def grab_data(self, Naverage=1, **kwargs):
        """
            | Start new acquisition.
            | grab the current values with keithley profile procedure.
            | Send the data_grabed_signal once done.

            =============== ======== ===============================================
            **Parameters**  **Type**  **Description**
            *Naverage*      int       Number of values to average
            =============== ======== ===============================================
        """
        data = [np.array([self.controller.get_power()])]
        self.data_grabed_signal.emit([DataFromPlugins(name='Powermeter', data=data,
                                                      dim='Data0D', labels=['Power (W)'],)])


    def stop(self):
        """
        """
        return ""


if __name__ == '__main__':
    main(__file__)



from pymodaq.control_modules.move_utility_classes import DAQ_Move_base
from pymodaq.control_modules.move_utility_classes import comon_parameters
from pymodaq.daq_utils.daq_utils import ThreadCommand, getLineInfo
from easydict import EasyDict as edict

import pylablib.devices.Thorlabs as Thorlabs

is_multiaxes = False
stage_names = []

class DAQ_Move_MFF101_pylablib(DAQ_Move_base):
    """
        Wrapper object to access the Flipper functionalities, similar wrapper for all controllers.
        =============== ==============
        **Attributes**    **Type**
        *params*          dictionnary
        =============== ==============
    """
    _controller_units = 'binary position'

    _dvc = Thorlabs.list_kinesis_devices()
    serialnumbers = [d[0] for d in _dvc if d[1] == 'APT Filter Flipper']

    params= [{'title': 'Controller ID:', 'name': 'controller_id', 'type': 'str', 'value': '', 'readonly': True},
             {'title': 'Serial number:', 'name': 'serial_number', 'type': 'list', 'limits': serialnumbers},
             {'title': 'Home Position:', 'name': 'home_position', 'type': 'list' , 'value': 0, 'limits' : [0,1]},
             {'title': 'MultiAxes:', 'name': 'multiaxes', 'type': 'group', 'visible': is_multiaxes, 'children':[
                        {'title': 'is Multiaxes:', 'name': 'ismultiaxes', 'type': 'bool', 'value': is_multiaxes, 'default': False},
                        {'title': 'Status:', 'name': 'multi_status', 'type': 'list', 'value': 'Master', 'limits': ['Master', 'Slave']},
                        {'title': 'Axis:', 'name': 'axis', 'type': 'list', 'limits': stage_names},
                        ]
              }
             ]+comon_parameters


    def __init__(self, parent=None, params_state=None):
        super().__init__(parent, params_state)
        self.settings.child('epsilon').setValue(0.1)
        self.settings.child('epsilon').setReadonly()

        self.settings.child('bounds', 'is_bounds').setValue(True)
        self.settings.child('bounds', 'is_bounds').setReadonly()

        self.settings.child('bounds', 'max_bound').setValue(1)
        self.settings.child('bounds', 'max_bound').setReadonly()

        self.settings.child('bounds', 'min_bound').setValue(0)
        self.settings.child('bounds', 'min_bound').setReadonly()

        self.settings.child('scaling','use_scaling').setValue(False)
        self.settings.child('scaling','use_scaling').setReadonly()
        self.settings.child('scaling','scaling').setValue(1)
        self.settings.child('scaling','scaling').setReadonly()
        self.settings.child('scaling','offset').setValue(0)
        self.settings.child('scaling','offset').setReadonly()

    def commit_settings(self,param):
        """
            | Activate any parameter changes on the hardware.
            | Called after a param_tree_changed signal received from DAQ_Move_main.

            =============== ================================ ========================
            **Parameters**  **Type**                          **Description**
            *param*         instance of pyqtgraph Parameter  The parameter to update
            =============== ================================ ========================
        """
        pass

    def ini_stage(self,controller=None):
        """Initialize the controller and stages (axes) with given parameters.

            ============== =========================================== ===========================================================================================
            **Parameters**  **Type**                                     **Description**

            *controller*    instance of the specific controller object  If defined this hardware will use it and will not initialize its own controller instance
            ============== =========================================== ===========================================================================================

            Returns
            -------
            Easydict
                dictionnary containing keys:
                 * *info* : string displaying various info
                 * *controller*: instance of the controller object in order to control other axes without the need to init the same controller twice
                 * *stage*: instance of the stage (axis or whatever) object
                 * *initialized*: boolean indicating if initialization has been done corretly

            See Also
            --------
            daq_utils.ThreadCommand
        """
        try:
            self.status.update(edict(info="",controller=None,initialized=False))

            #check whether this stage is controlled by a multiaxe controller (to be defined for each plugin)

            # if multiaxes then init the controller here if Master state otherwise use external controller
            if self.settings.child('multiaxes','ismultiaxes').value() and self.settings.child('multiaxes','multi_status').value()=="Slave":
                if controller is None: 
                    raise Exception('no controller has been defined externally while this axe is a slave one')
                else:
                    self.controller = controller
            else: #Master stage
                try:
                    controller = Thorlabs.kinesis.MFF(self.settings.child('serial_number').value())
                except:
                    raise Exception(f'No controller found with serial number {self.settings.child("serial_number").value()}')
                else:
                    self.controller = controller

            if not self.controller.is_opened():
                self.controller.open()

            #Getting the information from the device
            info = self.controller.get_device_info()

            #Setting the name of the controller == Type of controller
            self.settings.child('controller_id').setValue(info.notes)

            #Setting the transit time
            ttime_s = self.controller.get_flipper_parameters().transit_time
            self.settings.child('timeout').setValue(ttime_s*1e3) #Transit time ms

            #IDK what that does but ok
            self.status.info = info.notes
            self.status.controller = self.controller
            self.status.initialized = True
            return self.status
        #This is in case something fails
        except Exception as e:
            self.emit_status(ThreadCommand('Update_Status',[getLineInfo()+ str(e),'log']))
            self.status.info=getLineInfo()+ str(e)
            self.status.initialized=False
            return self.status

    def close(self):
        """
            close the current instance of Kinesis Flipper instrument.
        """
        self.controller.close()
        self.controller=None

    def stop_motion(self):
        """
            See Also
            --------
            DAQ_Move_base.move_done
        """
        # self.controller.stop(0)
        self.move_done()

    def check_position(self):
        """
            Get the current hardware position with scaling conversion of the Kinesis insrument provided by get_position_with_scaling

            See Also
            --------
            DAQ_Move_base.get_position_with_scaling, daq_utils.ThreadCommand
        """
        #Get position = 0 or 1, possibly None if unknown
        pos = self.controller.get_state()
        #Repollif pos returns none
        while pos == None:
            pos = self.controller.get_state()

        #This is superfluous as there is not scaling
        # pos = self.get_position_with_scaling(pos)
        self.emit_status(ThreadCommand('check_position', [pos]))
        return pos

    def move_Abs(self,position):
        """
            Make the hardware absolute move from the given position after thread command signal was received in DAQ_Move_main.

            =============== ========= =======================
            **Parameters**  **Type**   **Description**

            *position*       int       either 1 or 2 for the flipper
            =============== ========= =======================

            See Also
            --------
            DAQ_Move_base.set_position_with_scaling

        """
        position = self.check_bound(position)
        # superfluous because there is no scaling for a flipper
        # position = self.set_position_with_scaling(position)  # apply scaling if the user specified one

        # pos = self.check_position()
        # if pos == 0:
        #     position = 1
        # else:
        #     position = 0

        self.controller.move_to_state(position)
        self.emit_status(ThreadCommand('Update_Status', [f'Moving to position: {position}']))

        self.target_position = position
        self.poll_moving()


    def move_Rel(self,position):
        """ Move the actuator to the relative target actuator value defined by position
        Parameters
        ----------
        position: (float) value of the relative target positioning
        """
        self.current_position = self.check_position()
        # rel_move  = self.check_bound(self.current_position+position)-self.current_position
        new_pos = self.check_bound(self.current_position + position)
        self.target_position = new_pos

        # pos = self.check_position()
        # if pos == 0:
        #     position = 1
        # else:
        #     position = 0

        # self.target_position = position
        self.controller.move_to_state(new_pos)
        self.emit_status(ThreadCommand('Update_Status', [f'Moving to position: {position}']))
        self.poll_moving()


    def move_Home(self):
        """
            Make the absolute move to original position (0).
        """
        home = self.settings.child('home_position').value()
        self.target_position = home
        self.controller.move_to_state(home)
        self.emit_status(ThreadCommand('Update_Status', [f'Moving home: {home}']))
        self.poll_moving()


"""
PyMoDAQ plugin for thorlabs instruments based on the TLPM library allowing
remote control and monitoring of up to eight power and energy meters.
This software is compatible with our Power Meter Consoles and Interfaces (PM100A and PM101 Series),
Power and Energy Meter Consoles and Interfaces (PM100D, PM400, PM100USB, PM103 Series, and legacy PM200),
Position & Power Meter Interfaces (PM102 Series),
Wireless Power Meters (PM160, PM160T, and PM160T-HP),
and USB-Interface Power Meters (PM16 Series)

you have to install the Optical Monitor Software from Thorlabs to obtain the library

The installation should create (following the manual) an environment variable called either VXIPNPPATH64 or
VXIPNPPATH depending on your platform (32 or 64 bits) pointing to where the TLPM library is
(usually C:\Program Files\IVI Foundation\VISA)

This plugin is using the Instrumental_lib package directly interfacing the C library with the nice **instrument** wrapper
"""
import sys
from qtpy.QtCore import QThread
from easydict import EasyDict as edict
from pymodaq.utils.daq_utils import ThreadCommand, getLineInfo
from pymodaq.utils.data import DataFromPlugins
from pymodaq.control_modules.viewer_utility_classes import DAQ_Viewer_base, main
from collections import OrderedDict
import numpy as np
from pymodaq.control_modules.viewer_utility_classes import comon_parameters

from instrumental import list_instruments, instrument, Q_


psets = list_instruments(module='powermeters.thorlabs_tlpm')
DEVICES = [f"{pset['model']}/{pset['serial']}" for pset in psets]


class DAQ_0DViewer_TLPMPowermeterInst(DAQ_Viewer_base):

    _controller_units = 'W'

    params = comon_parameters + [
        {'title': 'Devices:', 'name': 'devices', 'type': 'list', 'limits': DEVICES},
        {'title': 'Info:', 'name': 'info', 'type': 'str', 'value': '', 'readonly': True},
        {'title': 'Wavelength:', 'name': 'wavelength', 'type': 'float', 'value': 532., 'suffix': 'nm'},
        ]

    def __init__(self, parent=None, params_state=None):
        super().__init__(parent, params_state)

    def ini_detector(self, controller=None):
        self.controller = self.ini_detector_init(controller, instrument(psets[DEVICES.index(self.settings['devices'])]))
        self.controller.power_unit = self._controller_units

        info = str(self.controller.get_device_info())

        self.settings.child('info').setValue(str(info))
        self.settings.child('wavelength').setOpts(
            limits=[wrange.magnitude for wrange in self.controller.get_wavelength_range()])
        self.settings.child('wavelength').setValue(self.controller.wavelength.magnitude)

        initialized = True
        return info, initialized

    def commit_settings(self, param):
        """
        """
        if param.name() == 'wavelength':
            self.controller.wavelength = Q_(self.settings.child('wavelength').value(), 'nm')
            self.settings.child('wavelength').setValue(self.controller.wavelength.m_as('nm'))

    def close(self):
        """
            close the current instance of Keithley viewer.
        """
        self.controller.close()

    def grab_data(self, Naverage=1, **kwargs):
        """
            | Start new acquisition.
            | grab the current values with keithley profile procedure.
            | Send the data_grabed_signal once done.

            =============== ======== ===============================================
            **Parameters**  **Type**  **Description**
            *Naverage*      int       Number of values to average
            =============== ======== ===============================================
        """
        data = np.array([np.mean([self.controller.get_power().magnitude for ind in range(Naverage)])])
        self.data_grabed_signal.emit([DataFromPlugins(name='Powermeter', data=[data],
                                                      dim='Data0D', labels=['Power (W)'],)])


    def stop(self):
        """
        """
        return ""


if __name__ == '__main__':
    main(__file__)



from pymodaq.control_modules.move_utility_classes import DAQ_Move_base, main, comon_parameters_fun
from pymodaq.utils.logger import set_logger, get_module_name

from pymodaq_plugins_thorlabs.hardware.kinesis import serialnumbers_flipper, Flipper

logger = set_logger(get_module_name(__file__))


class DAQ_Move_KinesisFlipper(DAQ_Move_base):
    """

    """
    _controller_units = 'degrees'
    _epsilon = 0.05

    is_multiaxes = False

    stage_names = []

    params = [{'title': 'Controller ID:', 'name': 'controller_id', 'type': 'str', 'value': '', 'readonly': True},
              {'title': 'Serial number:', 'name': 'serial_number', 'type': 'list',
               'limits': serialnumbers_flipper},
              ] + comon_parameters_fun(is_multiaxes, epsilon=_epsilon)

    def ini_attributes(self):
        self.controller: Flipper = None
        self.settings.child('bounds', 'is_bounds').setValue(True)
        self.settings.child('bounds', 'max_bound').setValue(1)
        self.settings.child('bounds', 'min_bound').setValue(0)

    def commit_settings(self, param):
        if param.name() == 'backlash':
            self.controller.backlash = param.value()

    def ini_stage(self, controller=None):
        """
        """
        self.controller = self.ini_stage_init(controller, Flipper())

        if self.settings['multiaxes', 'multi_status'] == "Master":
            self.controller.connect(self.settings['serial_number'])

        info = self.controller.name
        self.settings.child('controller_id').setValue(info)

        initialized = True
        return info, initialized

    def close(self):
        """
            close the current instance of Kinesis instrument.
        """
        self.controller.close()

    def stop_motion(self):
        """
            See Also
            --------
            DAQ_Move_base.move_done
        """
        self.controller.stop()

    def get_actuator_value(self):
        """
            Get the current hardware position with scaling conversion of the Kinsesis insrument provided by get_position_with_scaling

            See Also
            --------
            DAQ_Move_base.get_position_with_scaling, daq_utils.ThreadCommand
        """

        pos = self.controller.get_position()
        pos = self.get_position_with_scaling(pos)
        return pos

    def move_abs(self, position):
        """


        """
        position = self.check_bound(position)
        self.target_position = position
        position = self.set_position_with_scaling(position)

        self.controller.move_abs(position)

    def move_rel(self, position):
        """

        """
        position = self.check_bound(self.current_position + position) - self.current_position
        self.target_position = position + self.current_position
        position = self.set_position_relative_with_scaling(position)

        self.controller.move_rel(position)

    def move_home(self):
        """
        """
        self.controller.home(callback=self.move_done)


if __name__ == '__main__':
    main(__file__, init=False)



from easydict import EasyDict as edict
from pymodaq.utils.daq_utils import ThreadCommand, getLineInfo
from pymodaq.utils.data import DataFromPlugins, Axis
from pymodaq.control_modules.viewer_utility_classes import DAQ_Viewer_base, comon_parameters, main
from instrumental import instrument, list_instruments, Q_

# This is a (probably bad) way of importing the stuff needed to get exposure range
import instrumental.drivers.cameras.uc480 as uc480module


class DAQ_2DViewer_Thorlabs_DCx(DAQ_Viewer_base):
    """This plugin is intended for Thorlabs DCx cameras series.
        It should not be compatible with Thorlabs scientific cameras.

        This plugin use the instrumental library:
        https://instrumental-lib.readthedocs.io/en/stable/

        The class we use is defined here:
        https://github.com/mabuchilab/Instrumental/blob/master/instrumental/drivers/
        cameras/uc480.py

        Prerequisite
        ------------
        This plugin works with Windows 10.
        Installation procedure can be found here:
            https://instrumental-lib.readthedocs.io/en/stable/uc480-cameras.html
        In particular, the ThorCam software should be installed and the .dll libraries
            folder (where you can find uc480_64.dll) should be added in the environment
            PATH variable.
        In principle the dependencies (pywin32, nicelib) should be installed
            automatically while installing the plugin.
    """

    # Look for plugged cameras and get the serial numbers
    plugged_cameras = list_instruments(module='cameras.uc480')
    serial_numbers = []
    for paramset in plugged_cameras:
        camera = instrument(paramset, reopen_policy='reuse')
        serial_numbers.append(camera.serial.decode("utf-8"))

    params = comon_parameters + [
        {'title': 'Serial number:', 'name': 'serial_number', 'type': 'list', 'limits': serial_numbers},
        {'title': 'Exposure (ms):', 'name': 'exposure', 'type': 'float', 'value': 0},
        {'title': 'Gain:', 'name': 'master_gain', 'type': 'int', 'value': 0, "limits": [0, 100]},
        {'title': 'Gain Boost:', 'name': 'gain_boost', 'type': 'bool', 'value': False},
        {'title': 'gamma', 'name': 'gamma', 'type': 'int', 'value': 0},
        {'title': 'Color Mode', 'name': 'colormode', 'type': 'str', 'value': 'mono8', "readonly": True},
    ]

    def __init__(self, parent=None, params_state=None):
        super().__init__(parent, params_state)

        self.x_axis = None
        self.y_axis = None

        self.controller = None

    def commit_settings(self, param):
        """
        """
        if param.name() == 'exposure':
            # For some reason exposure is dealt specially in the instrumental lib
            self.controller._set_exposure(Q_(param.value(), 'ms'))
            self.settings.child('exposure').setValue(self.controller._get_exposure().m_as('ms'))
        elif param.name() in ['master_gain', 'gain_boost', 'gamma']:
            # All settings without units can be dealt as a single case
            setattr(self.controller, param.name(), param.value())
            self.settings.child(param.name()).setValue(getattr(self.controller, param.name()))

    def ini_detector(self, controller=None):
        """Detector communication initialization
        Parameters
        ----------
        controller: (object) custom object of a PyMoDAQ plugin (Slave case).
            None if only one detector by controller (Master case)
        Returns
        -------
        self.status (edict): with initialization status: three fields:
            * info (str)
            * controller (object) initialized controller
            * initialized: (bool): False if initialization failed otherwise True
        """

        try:
            self.status.update(edict(initialized=False, info="", x_axis=None,
                                     y_axis=None, controller=None))
            if self.settings.child('controller_status').value() == "Slave":
                if controller is None:
                    raise Exception('no controller has been defined externally while'
                                    'this detector is a slave one')
                else:
                    self.controller = controller
            else:
                camera_serial = self.settings.child('serial_number').value()
                plugged_cameras = list_instruments(module='cameras.uc480')
                selected_camera = None
                # Find the paramset that has the selected serial number
                for paramset in plugged_cameras:
                    camera = instrument(paramset, reopen_policy='reuse')
                    if camera.serial.decode("utf-8") == camera_serial:
                        selected_camera = camera

                self.controller = selected_camera

                # Getting the current settings from the instrument.
                # Exposure is weird
                self.settings.child('exposure').setValue(self.controller._get_exposure().m_as('ms'))
                for paraname in ['master_gain', 'gain_boost', 'gamma']:
                    paramvalue = getattr(self.controller, paraname)
                    self.settings.child(paraname).setValue(paramvalue)

                # Getting the range of exposure possible. I think it changes with other settings of the camera so it's
                # probably not ideal to set it like this. The _dev calls are needed because instrumental does not
                # natively exposes these parameters like it does e.g. for _get_exposure()

                rangemin = self.controller._dev.Exposure(uc480module.lib.IS_EXPOSURE_CMD_GET_EXPOSURE_RANGE_MIN)
                rangemax = self.controller._dev.Exposure(uc480module.lib.IS_EXPOSURE_CMD_GET_EXPOSURE_RANGE_MAX)
                self.settings.child('exposure').setOpts(limits=[rangemin, rangemax])

            self.status.info = "Detector initialized"
            self.status.initialized = True
            self.status.controller = self.controller
            return self.status

        except Exception as e:
            self.emit_status(
                ThreadCommand('Update_Status', [getLineInfo() + str(e), 'log']))
            self.status.info = getLineInfo() + str(e)
            self.status.initialized = False
            return self.status

    def close(self):
        """
        Terminate the communication protocol
        """
        self.controller.close()

    def grab_data(self, Naverage=1, **kwargs):
        """
        Parameters
        ----------
        Naverage: (int) Number of hardware averaging
        kwargs: (dict) of others optionals arguments
        """
        #The instrumental library seems really broken AF unfortunately so I have to use this to acquire a frame otherwise
        #it just uses the default values...
        kwds = {'exposure_time': Q_(self.settings.child('exposure').value(), 'ms'),
                'gain': self.settings.child('master_gain').value()}

        data = self.controller.grab_image(**kwds)

        if len(data.shape) > 2:
            data_list = [data[..., ind] for ind in range(data.shape[2])]
        else:
            data_list = [data]

        # data = self.controller.grab_image(exposure_time=Q_(self.settings.child('exposure').value(), 'ms'))
        self.data_grabed_signal.emit([DataFromPlugins(name='Thorcam', data=data_list,
                                                      dim='Data2D')])

    def stop(self):

        self.controller.stop_live_video()
        # self.emit_status(ThreadCommand('Update_Status', ['Some info you want to log']))
        return ''


if __name__ == '__main__':
    main(__file__, init=False)